package com.streambase.contrib.fitting;

import java.util.ArrayList;
import java.util.List;
import JSci.maths.*;
import JSci.maths.matrices.AbstractDoubleSquareMatrix;
import JSci.maths.matrices.DoubleMatrix;
import JSci.maths.polynomials.RealPolynomial;
import JSci.maths.statistics.FDistribution;
import JSci.maths.statistics.TDistribution;
import JSci.maths.vectors.AbstractDoubleVector;
import JSci.maths.vectors.DoubleVector;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.NullValueException;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.client.CustomFunctionResolver;

public class LeastSquaresFit {
	
	/**
	 * This function performs returns the following regression statistics:
	 *	Coefficients:		The coefficients associated with the regression performed in the following order 
	 *							(beta0 + beta1*x + beta2*x^2 + .....)
	 *	StandardErrors:		The standard errors associated with each of the coefficients
	 *	tStatistic:			The t-statistic associated with each regression coefficient
	 *	tpValue:			The p-value to test whether the coefficient is significant (whether the coefficient differs significantly from zero) 
	 *	Variance:			The model variance or the residual variance
	 *	RSquared:			The r- squared associated with the model (the explained variance of the model)
	 *	AdjustedRSquared:	A modification of the r-squared statistic  that adjusts for the number of explanatory terms in a model 
	 *						The adjusted R2 increases only if new terms improve the model more than would be expected by chance)
	 *	fStatistic:			The f-statistic associated with the joint probability of all the coefficients being zero
	 *	fpValue:			The p-value to test whether the fStatistic is significant
	 *	 	
	 * @param array The endogenous variables of the regression
	 * @param result The regression coefficients 
	 * @param y The exogenous (or explained) variables associated with the regression
	 * @return A tuple representing the regression statistics as listed above
	 */	
	public static Tuple fitStats(double [][] array, double [] yValues, AbstractDoubleVector result){	
		// A 2-dimensional matrix representing the X-values 
		// from the data (including a row of ones for the intercept)
		DoubleMatrix matrix = new DoubleMatrix(array);		
		// Transpose the matrix to get rows of data and columns 
		// associated with each regression coefficient	
		DoubleMatrix matrixPrime = (DoubleMatrix) matrix.transpose();		
		// The vector of y-values from the model 
		AbstractDoubleVector f = matrix.multiply(result);	
		// The vector of y-values from the data
		DoubleVector y = new DoubleVector(yValues);		
		// The residual (difference between the dat and the model)
		AbstractDoubleVector residual =  y.subtract(f);
		// The number of rows represents the sample size
		int n = residual.dimension();
		// The number of regressors in the linear model not counting the constant term
		int k = result.dimension();			
		// Create the distributions to use for signficance tests
		FDistribution FDist = new FDistribution(k-1,n-k);
	    TDistribution TDist = new TDistribution(n-k);

		// The variance-covariance matrix
		AbstractDoubleSquareMatrix varMatrix = (AbstractDoubleSquareMatrix) (matrixPrime.multiply(matrix));
		AbstractDoubleSquareMatrix varcov = (AbstractDoubleSquareMatrix) varMatrix.inverse();
				
		// Compute the mean of the y-values from the data set
		double yMean = ArrayMath.mean(yValues);

		// Compute the total sum of squares
		double SSTotal = 0,SSError = Math.pow(residual.norm(),2);
		
		// The residual variance (the error associated with the model)
		double modelVariance = SSError/(n-k);
		for (int i =0; i < y.dimension(); i++){
			SSTotal += Math.pow(y.getComponent(i)-yMean, 2);
		}
			
		// Compute the r-squared and the adjusted r-squared
		double rSquared = 1- SSError/SSTotal;
		double adjustedRsquared = rSquared -(1-rSquared)*(k-1)/(n - k);
					    
	    // Creating fields to add to the tuple
		Field [] field = new Field[9];
		field[0] = Schema.createListField("Coefficients", CompleteDataType.forDouble());
		field[1] = Schema.createListField("StandardErrors", CompleteDataType.forDouble());
		field[2] = Schema.createListField("tStatistic", CompleteDataType.forDouble());
		field[3] = Schema.createListField("tpValue", CompleteDataType.forDouble());
		field[4] = Schema.createField(DataType.DOUBLE, "Variance");
		field[5] = Schema.createField(DataType.DOUBLE, "RSquared");
		field[6] = Schema.createField(DataType.DOUBLE, "AdjustedRSquared");
		field[7] = Schema.createField(DataType.DOUBLE, "fStatistic");
		field[8] = Schema.createField(DataType.DOUBLE, "fpValue");
		
		// Creating a schema
		Schema schema = new Schema(null, field);
		
		// Creating a tuple from the schema above
		Tuple tuple = schema.createTuple();

		// Return a list of its regression coefficients
		ArrayList<Double> coefficients = new ArrayList<Double>();
	    for (int i= 0; i < result.dimension(); i++){	    	
			coefficients.add(result.getComponent(i));	    	
		}	    
	    
	    // Calculate the regression errors, the t-values, and the p-values 
	    // corresponding to the t-statistic
		ArrayList<Double> coefficientErrors = new ArrayList<Double>();
		ArrayList<Double> t = new ArrayList<Double>();
		ArrayList<Double> tpValue = new ArrayList<Double>();
	    for (int i = 0; i < result.dimension(); i++){	    	
			coefficientErrors.add(Math.sqrt(modelVariance * varcov.getElement(i,i)));
			t.add(coefficients.get(i)/coefficientErrors.get(i));
			tpValue.add(2*TDist.cumulative(-Math.abs(t.get(i))));
		}	    
	    
	    // The F-statistic and its corresponding p-value
	    double F = ((SSTotal-SSError)/(k-1))/(SSError/(n-k));
	    double fpValue = 1-FDist.cumulative(F);
	    
		// Write the values obtained to the tuple
		try {			
			tuple.setList("Coefficients", coefficients);
			tuple.setList("StandardErrors", coefficientErrors);
			tuple.setList("tStatistic", t);
			tuple.setList("tpValue", tpValue);
			tuple.setDouble("Variance", modelVariance);
			tuple.setDouble("RSquared", rSquared);
			tuple.setDouble("AdjustedRSquared", adjustedRsquared);
			tuple.setDouble("fStatistic", F);
			tuple.setDouble("fpValue", fpValue);
		} catch (NullValueException e) {
			System.err.println("Null value passed into a tuple field");
			e.printStackTrace();
		} catch (TupleException e) {
			System.err.println("An error was generated creating a tuple");
			e.printStackTrace();
		}
		return tuple;
	}

	/**
	 * This function fits a n degree polynomial to a dataset	 
	 * @param xdata The data from the independent variable
	 * @param ydata	The data from the dependent variable			
	 * @param degree The highest degree of the polynomial that will be fitted
	 * @return A tuple representing the coefficients of the regression operation  of the form:
	 * 		   y(x1, x2, ...) = coeffs(0) + coeffs(1) * x + coeffs(2) * x^2 + ...
	 * 		   This function also returns other fields in the tuple that measure how good the fit is (see fitStats)
	 */	
	@CustomFunctionResolver("RegressionResolver")
	public static Tuple leastSquares(List<?> xdata, List<?> ydata, int degree) throws StreamBaseException{
		// Declare variables and arrays that will be used in program		
		if (xdata.isEmpty()|| ydata.isEmpty() || degree < 1)
			return null;
		int len = xdata.size();
		if(len < 2 || len != ydata.size() || degree >= len)
			return null;
		
		// Declare the appropriate variables
		ArrayList <Double> x = new ArrayList<Double>();
		double [][] matrix = new double [2][len];
		for(int i = 0; i < len; i ++){
			x.add((Double) xdata.get(i));
			matrix[0][i] = x.get(i).doubleValue();
			matrix[1][i] = ((Double) ydata.get(i)).doubleValue();
		}
		
		// Create the design matrix
		double [][] design = new double [len][degree+1];
		for(int i = 0; i < len; i++){
			for(int j = 0; j <= degree; j ++){
				design[i][j] = Math.pow(x.get(i).doubleValue(), j);				
			}		
		}		 
		
		// Perform polynomial linear regression
		RealPolynomial result = LinearMath.leastSquaresFit(degree, matrix);
		
		// Create the double vector
		AbstractDoubleVector coefficients = new DoubleVector(result.getCoefficientsAsDoubles());
		
		//Return the coefficients with regression statistics 	   
	    return fitStats(design, matrix[1],coefficients);
	}
	
	/**
	 * A StreamBase Custom Function Resolver Function. This method is used by 
	 * StreamBase for type checking.  Use this method
	 * to resolve functions that take lists or tuples as arguments 
	 * or return lists or tuples.  This method should be used to ensure the 
	 * data types of your function are correct.  
	 */
	public static CompleteDataType RegressionResolver(
			CompleteDataType xdata,CompleteDataType ydata,CompleteDataType degree) {
		if (degree.getDataType() != DataType.INT){return null;}
		if (xdata == null){return null;}
		if (ydata == null){return null;}
		CompleteDataType argX = xdata.getElementType();
		CompleteDataType argY = ydata.getElementType();			
		if(argX.getDataType() == DataType.DOUBLE && argY.getDataType() == DataType.DOUBLE){
			Field [] field = new Field[9];
			field[0] = Schema.createListField("Coefficients", CompleteDataType.forDouble());
			field[1] = Schema.createListField("StandardErrors", CompleteDataType.forDouble());
			field[2] = Schema.createListField("tStatistic", CompleteDataType.forDouble());
			field[3] = Schema.createListField("tpValue", CompleteDataType.forDouble());
			field[4] = Schema.createField(DataType.DOUBLE, "Variance");
			field[5] = Schema.createField(DataType.DOUBLE, "RSquared");
			field[6] = Schema.createField(DataType.DOUBLE, "AdjustedRSquared");
			field[7] = Schema.createField(DataType.DOUBLE, "fStatistic");
			field[8] = Schema.createField(DataType.DOUBLE, "fpValue");
			Schema schema = new Schema("Output", field);
			return CompleteDataType.forTuple(schema);
		}	
		return null;
	}
}
